"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
from typing import Any

from .base_layer import BaseLayer


class YoloLayer(BaseLayer):
    def __init__(self, index: int, type_index: int):
        super().__init__(index=index, type_index=type_index, type_name="yolo")
        self._anchors: np.ndarray
        self._beta_nms: float
        self._classes = 20
        self._cls_normalizer = 1.0
        self._iou_loss = "iou"
        self._iou_thresh = 1.0
        self._iou_normalizer = 0.75
        self._label_smooth_eps = 0.0
        self._mask: np.ndarray
        self._max = 200
        self._new_coords = False
        self._nms_kind = "greedynms"
        self._num = 1
        self._obj_normalizer = 1.0
        self._scale_x_y = 1.0

    @property
    def anchors(self) -> np.ndarray:
        return self._anchors

    @property
    def beta_nms(self) -> float:
        if self._nms_kind == "greedynms":
            return 0.6
        return self._beta_nms

    @property
    def bflops(self) -> float:
        return 0

    @property
    def classes(self) -> int:
        return self._classes

    @property
    def cls_normalizer(self) -> float:
        return self._cls_normalizer

    @property
    def iou_loss(self) -> str:
        return self._iou_loss

    @property
    def iou_thresh(self) -> float:
        """
        Recommended to use iou_thresh=0.213
        """
        return self._iou_thresh

    @property
    def iou_normalizer(self) -> float:
        return self._iou_normalizer

    @property
    def label_smooth_eps(self) -> float:
        return self._label_smooth_eps

    @property
    def mask(self) -> np.ndarray:
        return self._mask

    @property
    def max(self) -> int:
        return self._max

    @property
    def new_coords(self) -> bool:
        return self._new_coords

    @property
    def nms_kind(self) -> str:
        return self._nms_kind

    @property
    def obj_normalizer(self) -> float:
        return self._obj_normalizer

    @property
    def scale_x_y(self) -> float:
        return self._scale_x_y

    @property
    def total(self) -> int:
        return self._num

    def __repr__(self) -> str:
        rep = f"{self.index:4}  "
        rep += f"{self.type[:5]:5s}_"
        rep += f"{self.type_index:<3}  "
        rep += f"iou_loss: {self._iou_loss}, "
        rep += f"iou_norm: {self._iou_normalizer}, "
        rep += f"obj_norm: {self._obj_normalizer}, "
        rep += f"cls_norm: {self._cls_normalizer}, "
        rep += "\n                 "
        rep += f"scale_x_y: {self._scale_x_y}, "
        rep += f"new_coords: {self._new_coords}, "
        rep += f"NMS: {self._nms_kind}, "
        rep += f"beta_nms: {self._beta_nms}, "
        rep += "\n                 "
        rep += f"iou_thresh: {self._iou_thresh}, "
        rep += f"label_smooth_eps: {self._label_smooth_eps}, "
        return rep

    def __setitem__(self, key: str, value: Any):
        if key in (
            "iou_loss",
            "nms_kind",
        ):
            self.__setattr__(f"_{key}", str(value))
        elif key in ("classes", "max", "num"):
            self.__setattr__(f"_{key}", int(value))
        elif key in (
            "beta_nms",
            "cls_normalizer",
            "iou_thresh",
            "iou_normalizer",
            "label_smooth_eps",
            "obj_normalizer",
            "scale_x_y",
        ):
            self.__setattr__(f"_{key}", float(value))
        elif key in ("new_coords",):
            self.__setattr__(f"_{key}", bool(int(value)))
        elif key in ("mask",):
            self.__setattr__(
                f"_{key}",
                np.array([int(i.strip()) for i in value.split(",")], np.int32),
            )
        elif key == "anchors":
            value = [int(i.strip()) for i in value.split(",")]
            _value = []
            for i in range(len(value) // 2):
                _value.append((value[2 * i], value[2 * i + 1]))
            self.__setattr__(f"_{key}", np.array(_value, np.float32))
        elif key == "input_shape":
            self.__setattr__(f"_{key}", value)
            self._output_shape = self._input_shape
        else:
            raise KeyError(f"'{key}' is not supported")



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.losses import Loss

from .iou import bbox_iou, bbox_ciou, bbox_giou
from ..model import YOLOv4Model
from ...common.config import YOLOConfig

_BBOX_XIOU_MAP = {
    "iou": bbox_iou,
    "ciou": bbox_ciou,
    "giou": bbox_giou,
}


class YOLOv4Loss(Loss):
    def __init__(self, config: YOLOConfig, model: YOLOv4Model):
        super().__init__(name="YOLOv4Loss")
        self._metayolos = []
        for i in range(config.layer_count["yolo"]):
            self._metayolos.append(config.find_metalayer("yolo", i))
        self._metanet = config.net
        self.model = model

        self._bbox_xiou = _BBOX_XIOU_MAP[self._metayolos[-1].iou_loss]
        self._num_mask = tf.constant(
            len(self._metayolos[-1].mask), dtype=tf.int32
        )

        # yolo #################################################################

        self._iou_norm = tf.constant(
            [metayolo.iou_normalizer for metayolo in self._metayolos],
            dtype=tf.float32,
        )
        self._cls_norm = tf.constant(
            [metayolo.cls_normalizer for metayolo in self._metayolos],
            dtype=tf.float32,
        )
        self._obj_norm = tf.constant(
            [metayolo.obj_normalizer for metayolo in self._metayolos],
            dtype=tf.float32,
        )

        # anchor ###############################################################

        stride = self._metayolos[-1].classes + 5
        self._box_index = tf.constant(
            [stride * i for i in range(len(self._metayolos[-1].mask))],
            dtype=tf.int32,
        )
        self._obj_index = tf.constant(
            [stride * i + 4 for i in range(len(self._metayolos[-1].mask))],
            dtype=tf.int32,
        )
        self._cls_index = tf.constant(
            [stride * i + 5 for i in range(len(self._metayolos[-1].mask))],
            dtype=tf.int32,
        )
        self._next_box_index = tf.constant(
            [stride * (i + 1) for i in range(len(self._metayolos[-1].mask))],
            dtype=tf.int32,
        )
        self._one_index = tf.constant(
            [
                self._metayolos[-1].channels + i
                for i in range(len(self._metayolos[-1].mask))
            ],
            dtype=tf.int32,
        )

    def call(self, y_true, y_pred):
        """
        @param `y_true`: Dim(batch, yolo.h, yolo.w, yolo.c)
        @param `y_pred`: Dim(batch, yolo.h, yolo.w, yolo.c + len(mask))
        """
        yolo_name = y_pred.name.split("/")[-2]
        yolo_index = int(yolo_name.split("_")[-1])

        def anchor_loop(anchor, iou_loss0, obj_loss0, cls_loss0):
            true_box = y_true[
                ..., self._box_index[anchor] : self._obj_index[anchor]
            ]
            true_obj = y_true[..., self._obj_index[anchor]]
            true_cls = y_true[
                ..., self._cls_index[anchor] : self._next_box_index[anchor]
            ]
            true_one = y_true[..., self._one_index[anchor]]

            pred_box = y_pred[
                ..., self._box_index[anchor] : self._obj_index[anchor]
            ]
            pred_obj = y_pred[..., self._obj_index[anchor]]
            pred_cls = y_pred[
                ..., self._cls_index[anchor] : self._next_box_index[anchor]
            ]

            # obj loss
            obj_loss1 = self._obj_norm[yolo_index] * K.sum(
                K.binary_crossentropy(true_obj, pred_obj)
            )

            # cls loss
            cls_loss1 = self._cls_norm[yolo_index] * K.sum(
                true_one
                * K.sum(
                    K.binary_crossentropy(true_cls, pred_cls),
                    axis=-1,
                )
            )

            xious, ious = self._bbox_xiou(pred_box, true_box)

            # xiou loss
            iou_loss1 = self._iou_norm[yolo_index] * K.sum(
                true_one * (1 - xious)
            )

            # metrics update
            ious = true_one * ious
            self.model._ious.assign_add(K.sum(ious))
            self.model._recall50.assign_add(
                K.sum(tf.cast(ious > 0.5, dtype=tf.int32))
            )
            self.model._recall75.assign_add(
                K.sum(tf.cast(ious > 0.75, dtype=tf.int32))
            )

            return (
                tf.add(anchor, 1),
                tf.add(iou_loss0, iou_loss1),
                tf.add(obj_loss0, obj_loss1),
                tf.add(cls_loss0, cls_loss1),
            )

        _, iou_loss0, obj_loss0, cls_loss0 = tf.while_loop(
            cond=lambda anchor, _, __, ___: tf.less(anchor, self._num_mask),
            body=anchor_loop,
            loop_vars=[0, 0.0, 0.0, 0.0],
        )

        self.model._total_truth.assign_add(
            tf.cast(K.sum(y_true[..., self._one_index[0] :]), dtype=tf.int64)
        )

        iou_loss0 /= self._metanet.batch
        obj_loss0 /= self._metanet.batch
        cls_loss0 /= self._metanet.batch
        total_loss = iou_loss0 + obj_loss0 + cls_loss0

        self.model._iou_loss.assign_add(iou_loss0)
        self.model._obj_loss.assign_add(obj_loss0)
        self.model._cls_loss.assign_add(cls_loss0)
        self.model._total_loss.assign_add(total_loss)

        return total_loss



"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import Any, Dict, List, Union

import numpy as np

from . import parser
from .metalayer import YoloLayer


class YOLOConfig:
    def __init__(self):
        self._metalayers: Dict[Union[str, int], Any] = {}
        self._layer_count: Dict[str, int]
        self._model_name: str
        self._names: Dict[int, str] = {}

    def find_metalayer(self, layer_type: str, layer_index: int) -> Any:
        """
        Usage:
            last_yolo_layer = config.find_metalayer("yolo", -1)
        """
        if layer_index < 0:
            count = self._layer_count[layer_type]
            layer_index = count + layer_index

        return self._metalayers[f"{layer_type}_{layer_index}"]

    def summary(self):
        print(self._metalayers[-1])
        print(
            "index layer No. filters  size/strd(dil)   input_shape         "
            "output_shape    1e9 flops"
        )
        total_bflops = 0
        for i in range(self._layer_count["total"]):
            print(self._metalayers[i])
            total_bflops += self._metalayers[i].bflops
        print(f"Total B(1e9)flops: {total_bflops:6.3f}")

    # Parse ####################################################################

    def parse_cfg(self, cfg_path: str):
        (
            self._metalayers,
            self._layer_count,
            self._model_name,
        ) = parser.parse_cfg(cfg_path=cfg_path)

        if self.layer_count["yolo"] > 0:
            self._metayolos = [
                self.find_metalayer("yolo", i)
                for i in range(self.layer_count["yolo"])
            ]
        elif self.layer_count["yolo_tpu"] > 0:
            self._metayolos = [
                self.find_metalayer("yolo_tpu", i)
                for i in range(self.layer_count["yolo_tpu"])
            ]

        if len(self._metayolos) > 0:
            self._masks = [yolo.mask for yolo in self._metayolos]

            self._anchors = self._metayolos[0].anchors / np.array(
                [self.net.width, self.net.height], np.float32
            )

    def parse_names(self, names_path: str):
        self._names = parser.parse_names(names_path=names_path)

    # Property #################################################################

    @property
    def anchors(self) -> np.ndarray:
        return self._anchors

    @property
    def layer_count(self) -> Dict[str, int]:
        """
        key: metalayer.type
        value: the number of layers of the same type
        """
        return self._layer_count

    @property
    def masks(self) -> List[np.ndarray]:
        return self._masks

    @property
    def metalayers(self) -> Dict[Union[str, int], Any]:
        return self._metalayers

    @property
    def metayolos(self) -> List[YoloLayer]:
        return self._metayolos

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def names(self) -> Dict[int, str]:
        """
        class names
        """
        return self._names

    # Magic ####################################################################

    def __getattr__(self, metalayer: str) -> Any:
        return self._metalayers[metalayer]



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np


def load_weights(model, weights_file: str):
    with open(weights_file, "rb") as fd:
        # major, minor, revision, seen, _
        _np_fromfile(fd, dtype=np.int32, count=5)

        for layer in model.layers:
            if "convolutional" in layer.name:
                if not yolo_conv2d_load_weights(layer, fd):
                    break

        if len(fd.read()) != 0:
            raise ValueError("Model and weights file do not match.")


def _np_fromfile(fd, dtype, count: int):
    data = np.fromfile(fd, dtype=dtype, count=count)
    if len(data) != count:
        if len(data) == 0:
            return None
        raise ValueError("Model and weights file do not match.")
    return data


def yolo_conv2d_load_weights(yolo_conv2d, fd) -> bool:
    conv2d = None
    batch_normalization = None
    for layer in yolo_conv2d.layers:
        if "batch_normalization" in layer.name:
            batch_normalization = layer
        elif "conv2d" in layer.name:
            conv2d = layer

    filters = conv2d.filters

    if batch_normalization is not None:
        # darknet weights: [beta, gamma, mean, variance]
        bn_weights = _np_fromfile(fd, dtype=np.float32, count=4 * filters)
        if bn_weights is None:
            return False
        # tf weights: [gamma, beta, mean, variance]
        bn_weights = bn_weights.reshape((4, filters))[[1, 0, 2, 3]]

        batch_normalization.set_weights(bn_weights)

    conv_bias = None
    if conv2d.use_bias:
        conv_bias = _np_fromfile(fd, dtype=np.float32, count=filters)
        if conv_bias is None:
            return False

    # darknet shape (out_dim, in_dim, kernel_size, kernel_size)
    conv_shape = (filters, conv2d.input_shape[-1], *conv2d.kernel_size)

    conv_weights = _np_fromfile(
        fd, dtype=np.float32, count=np.product(conv_shape)
    )
    if conv_weights is None:
        return False
    # tf shape (kernel_size, kernel_size, in_dim, out_dim)
    conv_weights = conv_weights.reshape(conv_shape).transpose([2, 3, 1, 0])

    if conv_bias is None:
        conv2d.set_weights([conv_weights])
    else:
        conv2d.set_weights([conv_weights, conv_bias])

    return True


def save_weights(model, weights_file: str, to: str = ""):
    with open(weights_file, "wb") as fd:
        # major, minor, revision, seen, _
        np.array([0, 2, 5, 32032000, 0], dtype=np.int32).tofile(fd)

        for layer in model.layers:
            if "convolutional" in layer.name:
                yolo_conv2d_save_weights(layer, fd)
                if layer.name == to:
                    break


def yolo_conv2d_save_weights(yolo_conv2d, fd):
    conv2d = None
    batch_normalization = None
    for layer in yolo_conv2d.layers:
        if "batch_normalization" in layer.name:
            batch_normalization = layer
        elif "conv2d" in layer.name:
            conv2d = layer

    if batch_normalization is not None:
        # tf weights: [gamma, beta, mean, variance]
        bn_weights = np.stack(batch_normalization.get_weights())
        # darknet weights: [beta, gamma, mean, variance]
        bn_weights[[1, 0, 2, 3]].reshape((-1,)).tofile(fd)

    # tf shape (height, width, in_dim, out_dim)
    if conv2d.use_bias:
        conv_weights, conv_bias = conv2d.get_weights()
        conv_bias.tofile(fd)
    else:
        conv_weights = conv2d.get_weights()[0]

    # darknet shape (out_dim, in_dim, height, width)
    conv_weights.transpose([3, 2, 0, 1]).reshape((-1,)).tofile(fd)



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import platform
from typing import List

import numpy as np

try:
    import tflite_runtime.interpreter as tflite
    from tflite_runtime.interpreter import load_delegate
except ModuleNotFoundError:
    import tensorflow.lite as tflite

    load_delegate = tflite.experimental.load_delegate

from ..common.base_class import BaseClass
from ..common import yolo_tpu_layer as _yolo_tpu_layer

EDGETPU_SHARED_LIB = {
    "Linux": "libedgetpu.so.1",
    "Darwin": "libedgetpu.1.dylib",
    "Windows": "edgetpu.dll",
}[platform.system()]


class YOLOv4(BaseClass):
    def load_tflite(
        self, tflite_path: str, edgetpu_lib: str = EDGETPU_SHARED_LIB
    ) -> None:
        self._tpu = self.config.layer_count["yolo_tpu"] > 0

        if self._tpu:
            self._interpreter = tflite.Interpreter(
                model_path=tflite_path,
                experimental_delegates=[load_delegate(edgetpu_lib)],
            )
        else:
            self._interpreter = tflite.Interpreter(model_path=tflite_path)

        self._interpreter.allocate_tensors()

        # input_details
        input_details = self._interpreter.get_input_details()[0]
        if (
            input_details["shape"][1] != self.config.net.input_shape[0]
            or input_details["shape"][2] != self.config.net.input_shape[1]
            or input_details["shape"][3] != self.config.net.input_shape[2]
        ):
            raise RuntimeError(
                "YOLOv4: config.input_shape and tflite.input_details['shape']"
                " do not match."
            )
        self._input_details = input_details
        self._input_float = self._input_details["dtype"] is np.float32

        # output_details
        self._output_details = self._interpreter.get_output_details()

        layer_type = ""
        if self._tpu:
            layer_type = "yolo_tpu"
        else:
            layer_type = "yolo"

        self._scale_x_y = []
        self._num_masks: int
        for i in range(self.config.layer_count[layer_type]):
            metayolo = self.config.find_metalayer(layer_type, i)
            self._scale_x_y.append(metayolo.scale_x_y)
            self._num_masks = len(metayolo.mask)

    def summary(self):
        self.config.summary()

    #############
    # Inference #
    #############

    def _predict(self, x: np.ndarray) -> List[np.ndarray]:
        self._interpreter.set_tensor(self._input_details["index"], x)
        self._interpreter.invoke()
        # [yolo0, yolo1, ...]
        # yolo == Dim(1, height, width, channels)
        # yolo_tpu == x, logistic(x)

        yolos = [
            self._interpreter.get_tensor(output_detail["index"])
            for output_detail in self._output_details
        ]

        if self._tpu:
            num_yolo = len(yolos) // 2
            _yolos = []
            for i in range(num_yolo):
                _yolo_tpu_layer(
                    yolos[2 * i],
                    yolos[2 * i + 1],
                    self._num_masks,
                    self._scale_x_y[i],
                )
                _yolos.append(yolos[2 * i + 1])

            return _yolos

        return yolos

    def predict(self, frame: np.ndarray) -> np.ndarray:
        """
        Predict one frame

        @param frame: Dim(height, width, channels)

        @return pred_bboxes
            Dim(-1, (x, y, w, h, cls_id, prob))
        """
        # image_data == Dim(1, input_size[1], input_size[0], channels)
        height, width, _ = frame.shape

        image_data = self.resize_image(frame)
        if self._input_float:
            candidates = self._predict(
                image_data[np.newaxis, ...].astype(np.float32) / 255
            )
        else:
            candidates = self._predict(image_data[np.newaxis, ...])

        # Select 0
        pred_bboxes = self.get_yolo_detections(yolos=candidates)
        self.fit_to_original(pred_bboxes, height, width)
        return pred_bboxes



"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import Any

from .base_layer import BaseLayer


class NetLayer(BaseLayer):
    def __init__(self, index: int, type_index: int):
        super().__init__(index=index, type_index=type_index, type_name="net")
        self._batch = 1
        self._burn_in = 0
        self._channels = 0
        self._height = 0
        self._learning_rate = 0.001
        self._max_batches = 0
        self._momentum = 0.9
        self._mosaic = False
        self._policy = "steps"
        self._power = 4
        self._scales: tuple
        self._steps: tuple
        self._width = 0

    @property
    def batch(self) -> int:
        return self._batch

    @property
    def burn_in(self) -> int:
        return self._burn_in

    @property
    def channels(self) -> int:
        # override
        return self._channels

    @property
    def height(self) -> int:
        # override
        return self._height

    @property
    def input_shape(self) -> tuple:
        # override
        return (self._height, self._width, self._channels)

    @property
    def learning_rate(self) -> float:
        return self._learning_rate

    @property
    def max_batches(self) -> int:
        return self._max_batches

    @property
    def momentum(self) -> float:
        return self._momentum

    @property
    def mosaic(self) -> bool:
        return self._mosaic

    @property
    def name(self) -> str:
        # override
        return self._type_

    @property
    def output_shape(self) -> tuple:
        # override
        return (self._height, self._width, self._channels)

    @property
    def policy(self) -> str:
        return self._policy

    @property
    def power(self) -> int:
        return self._power

    @property
    def scales(self) -> tuple:
        return self._scales

    @property
    def steps(self) -> tuple:
        return self._steps

    @property
    def width(self) -> int:
        # override
        return self._width

    def __repr__(self) -> str:
        rep = f"batch: {self._batch}"
        return rep

    def __setitem__(self, key: str, value: Any):
        if key in ("policy",):
            self.__setattr__(f"_{key}", str(value))
        elif key in (
            "batch",
            "burn_in",
            "channels",
            "height",
            "max_batches",
            "power",
            "width",
        ):
            self.__setattr__(f"_{key}", int(value))
        elif key in ("mosaic",):
            self.__setattr__(f"_{key}", bool(int(value)))
        elif key in ("learning_rate", "momentum"):
            self.__setattr__(f"_{key}", float(value))
        elif key in ("steps",):
            self.__setattr__(
                f"_{key}", tuple(int(i.strip()) for i in value.split(","))
            )
        elif key in ("scales",):
            self.__setattr__(
                f"_{key}", tuple(float(i.strip()) for i in value.split(","))
            )
        else:
            raise KeyError(f"'{key}' is not supported")



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Input
from tensorflow.keras.optimizers import Adam

from .dataset.keras_sequence import YOLODataset  # for exporting
from .model import YOLOv4Model
from .training.callbacks import (
    SaveWeightsCallback,  # for exporting
    YOLOCallbackAtEachStep,
)
from .training.yolo_loss import YOLOv4Loss
from .utils.mAP import create_mAP_input_files  # for exporting
from .utils.tflite import save_as_tflite  # for expoerting
from .utils.weights import (
    load_weights as _load_weights,
    save_weights as _save_weights,
)
from ..common.base_class import BaseClass

physical_devices = tf.config.experimental.list_physical_devices("GPU")
if len(physical_devices) > 0:
    tf.config.experimental.set_memory_growth(physical_devices[0], True)
    print("Call tf.config.experimental.set_memory_growth(GPU0, True)")


class YOLOv4(BaseClass):
    @property
    def model(self) -> YOLOv4Model:
        return self._model

    def make_model(self):
        K.clear_session()
        _input = Input(self.config.net.input_shape)
        self._model = YOLOv4Model(config=self.config)
        self._model(_input)

    def load_weights(self, weights_path: str, weights_type: str = "tf"):
        """
        Usage:
            yolo.load_weights("checkpoints")
            yolo.load_weights("yolov4.weights", weights_type="yolo")
        """
        if weights_type == "yolo":
            _load_weights(self._model, weights_path)
        elif weights_type == "tf":
            self._model.load_weights(weights_path)

    def save_weights(
        self, weights_path: str, weights_type: str = "tf", to: int = 0
    ):
        """
        Usage:
            yolo.save_weights("checkpoints")
            yolo.save_weights("yolov4.weights", weights_type="yolo")
            yolo.save_weights("yolov4.conv.137", weights_type="yolo", to=137)
        """
        to_layer = ""
        if to > 0:
            to_layer = self.config.metalayers[to - 1].name

        if weights_type == "yolo":
            _save_weights(self._model, weights_path, to=to_layer)
        elif weights_type == "tf":
            self._model.save_weights(weights_path)

    def summary(self, line_length=90, summary_type: str = "tf", **kwargs):
        if summary_type == "tf":
            self._model.summary(line_length=line_length, **kwargs)
        else:
            self.config.summary()

    #############
    # Inference #
    #############

    @tf.function
    def _predict(self, x):
        return self._model(x, training=False)

    def predict(self, frame: np.ndarray):
        """
        Predict one frame

        @param frame: Dim(height, width, channels)

        @return pred_bboxes
            Dim(-1, (x, y, w, h, cls_id, prob))
        """
        # image_data == Dim(1, input_size[1], input_size[0], channels)
        height, width, _ = frame.shape

        image_data = self.resize_image(frame)
        image_data = image_data / 255.0
        image_data = image_data[np.newaxis, ...].astype(np.float32)

        candidates = self._predict(image_data)
        candidates = [
            c.numpy().astype(np.float32, copy=False) for c in candidates
        ]

        # Select 0
        pred_bboxes = self.get_yolo_detections(yolos=candidates)
        self.fit_to_original(pred_bboxes, height, width)
        return pred_bboxes

    ############
    # Training #
    ############

    def compile(
        self,
        optimizer=None,
        loss=None,
        **kwargs,
    ):
        if optimizer is None:
            optimizer = Adam(learning_rate=self.config.net.learning_rate)

        if loss is None:
            loss = YOLOv4Loss(config=self.config, model=self.model)

        return self._model.compile(
            optimizer=optimizer,
            loss=loss,
            **kwargs,
        )

    def fit(
        self,
        dataset,
        callbacks=None,
        validation_data=None,
        validation_steps=None,
        verbose: int = 3,
        **kwargs,
    ):
        """
        verbose=3 is one line per step
        """
        callbacks = callbacks or []
        callbacks.append(
            YOLOCallbackAtEachStep(config=self.config, verbose=verbose)
        )

        epochs = self.config.net.max_batches // len(dataset) + 1

        return self._model.fit(
            dataset,
            epochs=epochs,
            verbose=verbose if verbose < 3 else 0,
            callbacks=callbacks,
            validation_data=validation_data,
            validation_steps=validation_steps,
            **kwargs,
        )



"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class BaseLayer:
    def __init__(self, index: int, type_index: int, type_name: str):
        self._index_ = index
        self._input_shape: tuple
        self._output_shape: tuple
        self._type_index_ = type_index
        self._type_ = type_name

    @property
    def channels(self) -> int:
        return self._input_shape[2]

    @property
    def height(self) -> int:
        return self._input_shape[0]

    @property
    def index(self) -> int:
        """
        in the total model
        """
        return self._index_

    @property
    def input_shape(self) -> tuple:
        """
        (height, width, channels)
        """
        return tuple(self._input_shape)

    @property
    def name(self) -> str:
        return f"{self._type_}_{self._type_index_}"

    @property
    def output_shape(self) -> tuple:
        return tuple(self._output_shape)

    @property
    def type_index(self) -> int:
        """
        in the same type
        """
        return self._type_index_

    @property
    def type(self) -> str:
        return self._type_

    @property
    def width(self) -> int:
        return self._input_shape[1]



"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from os import path
import pathlib
import random
from typing import Any, Dict, Tuple, Union

import numpy as np

from .metalayer import (
    ConvolutionalLayer,
    MaxpoolLayer,
    NetLayer,
    RouteLayer,
    ShortcutLayer,
    UpsampleLayer,
    YoloLayer,
    YoloTpuLayer,
)


def parse_cfg(
    cfg_path: str,
) -> Tuple[Dict[Union[str, int], Any], Dict[str, int], str]:
    """
    @return
        Dict[layer_name or layer_index, metalayer]
        Dict[layer_type, count]
        model_name
    """
    metalayers: Dict[Union[str, int], Any] = {}
    count: Dict[str, int] = {
        "convolutional": 0,
        "maxpool": 0,
        "net": 0,
        "route": 0,
        "shortcut": 0,
        "total": -1,
        "upsample": 0,
        "yolo": 0,
        "yolo_tpu": 0,
    }
    layer_type: str = "net"

    meta_layer: Dict[str, Any] = {
        "convolutional": ConvolutionalLayer,
        "maxpool": MaxpoolLayer,
        "net": NetLayer,
        "route": RouteLayer,
        "shortcut": ShortcutLayer,
        "upsample": UpsampleLayer,
        "yolo": YoloLayer,
        "yolo_tpu": YoloTpuLayer,
    }

    with open(cfg_path, "r") as fd:
        layer = NetLayer(index=-1, type_index=-1)
        for line in fd:
            line = line.strip().split("#")[0]
            if line == "":
                continue

            if line[0] == "[":
                layer_type = line[1:-1]
                count["total"] += 1
                count[layer_type] += 1

                layer = meta_layer[layer_type](
                    index=count["total"] - 1, type_index=count[layer_type] - 1
                )
                metalayers[layer.name] = layer
                metalayers[count["total"] - 1] = layer

            else:
                # layer option
                option, value = line.split("=")
                option = option.strip()
                value = value.strip()
                try:
                    metalayers[layer.name][option] = value
                except KeyError as error:
                    raise RuntimeError(
                        f"parse_cfg: [{layer.name}] '{option}' is not"
                        " supported."
                    ) from error

    # Build layer
    for index in range(count["total"]):
        layer = metalayers[index]

        output_shape = metalayers[index - 1].output_shape
        if layer.type in ("route", "shortcut"):
            if len(layer.layers) > 1:
                output_shape = [
                    metalayers[i].output_shape for i in layer.layers
                ]
            else:
                output_shape = metalayers[layer.layers[0]].output_shape
        layer["input_shape"] = output_shape

    model_name = pathlib.Path(cfg_path).stem

    return metalayers, count, model_name


def parse_names(names_path: str) -> Dict[int, str]:
    """
    @return {id: class name}
    """
    names: Dict[int, str] = {}
    with open(names_path, "r") as fd:
        index = 0
        for class_name in fd:
            class_name = class_name.strip()
            if len(class_name) != 0:
                names[index] = class_name
                index += 1

    return names


def parse_dataset(
    dataset_list: str,
    dataset_type: str = "converted_coco",
    image_path_prefix: str = "",
):
    """
    x: center x 0.0 ~ 1.0
    y: center y 0.0 ~ 1.0
    @return [
                [
                    image_path,
                    [
                        [x, y, w, h, class_id]
                        ,
                        ...
                    ]
                ],
                ...
            ]
    """
    dataset = []

    with open(dataset_list, "r") as fd:
        lines = fd.readlines()

        if dataset_type == "converted_coco":
            for line in lines:
                # line: "<image_path> class_id,x,y,w,h ..."
                bboxes = line.strip().split()

                image_path = bboxes[0]
                if image_path_prefix != "":
                    image_path = path.join(image_path_prefix, image_path)

                xywhc_s = np.zeros((len(bboxes) - 1, 5), dtype=np.float32)
                for i, bbox in enumerate(bboxes[1:]):
                    # bbox = class_id,x,y,w,h
                    bbox = list(map(float, bbox.split(",")))
                    xywhc_s[i, :] = (
                        *bbox[1:],
                        bbox[0],
                    )

                dataset.append([image_path, xywhc_s])

        elif dataset_type == "yolo":
            for line in lines:
                # line: "<image_path>"
                image_path = line.strip()
                if image_path_prefix != "":
                    image_path = path.join(image_path_prefix, image_path)

                root, _ = path.splitext(image_path)
                with open(root + ".txt") as fd2:
                    bboxes = fd2.readlines()
                    xywhc_s = np.zeros((len(bboxes), 5), dtype=np.float32)
                    for i, bbox in enumerate(bboxes):
                        # bbox = class_id x y w h
                        bbox = bbox.strip()
                        bbox = list(map(float, bbox.split(" ")))
                        xywhc_s[i, :] = (
                            *bbox[1:],
                            bbox[0],
                        )
                    dataset.append([image_path, xywhc_s])

    if len(dataset) == 0:
        raise RuntimeError(
            f"parse_dataset: There is no dataset in '{dataset_list}'."
        )

    # Select 5 sets randomly and check the data format
    for _ in range(5):
        first_bbox = dataset[random.randint(0, len(dataset) - 1)][1][0]
        for i in range(4):
            if first_bbox[i] < 0 or first_bbox[i] > 1:
                raise RuntimeError(
                    "parse_dataset: 'center_x', 'center_y', 'width', and"
                    " 'height' should be between 0.0 and 1.0."
                )

        if int(first_bbox[4]) < 0:
            raise RuntimeError(
                "parse_dataset: 'class_id' should be an integer greater than or"
                " equal to 0."
            )

    return dataset



"""
MIT License

Copyright (c) 2020-2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import colorsys
from typing import Dict, Union

# Don't import tensorflow
import cv2
import numpy as np


_HSV = [(1.0 * x / 256, 1.0, 1.0) for x in range(256)]
_COLORS = list(map(lambda x: colorsys.hsv_to_rgb(*x), _HSV))
_COLORS = list(
    map(
        lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)),
        _COLORS,
    )
)
BBOX_COLORS = []
_OFFSET = [0, 8, 4, 12, 2, 6, 10, 14, 1, 3, 5, 7, 9, 11, 13, 15]
for i in range(256):
    BBOX_COLORS.append(_COLORS[(i * 16) % 256 + _OFFSET[(i * 16) // 256]])


def resize_image(
    image: np.ndarray,
    target_shape: Union[list, tuple],
    ground_truth: np.ndarray = None,
):
    """
    @param `image`:        Dim(height, width, channels)
    @param `target_shape`:  (height, width, ...)
    @param `ground_truth`: [[center_x, center_y, w, h, class_id], ...]

    @return resized_image or (resized_image, resized_ground_truth)

    Usage:
        image = media.resize_image(image, yolo.input_size)
        image, ground_truth = media.resize_image(image, yolo.input_size,
                                                                ground_truth)
    """
    height, width, _ = image.shape
    target_height = target_shape[0]
    target_width = target_shape[1]

    if width / height >= target_width / target_height:
        scale = target_width / width
    else:
        scale = target_height / height

    # Resize
    if scale != 1:
        width = int(round(width * scale))
        height = int(round(height * scale))
        resized_image = cv2.resize(image, (width, height))
    else:
        resized_image = np.copy(image)

    # Pad
    dw = target_width - width
    dh = target_height - height

    if not (dw == 0 and dh == 0):
        dw = dw // 2
        dh = dh // 2
        # height, width, channels
        padded_image = np.full(
            (target_height, target_width, 3), 255, dtype=np.uint8
        )
        padded_image[dh : height + dh, dw : width + dw, :] = resized_image
    else:
        padded_image = resized_image

    if ground_truth is None:
        return padded_image

    # Resize ground truth
    ground_truth = np.copy(ground_truth)

    if dw > dh:
        scale = width / target_width
        ground_truth[:, 0] = scale * (ground_truth[:, 0] - 0.5) + 0.5
        ground_truth[:, 2] = scale * ground_truth[:, 2]
    elif dw < dh:
        scale = height / target_height
        ground_truth[:, 1] = scale * (ground_truth[:, 1] - 0.5) + 0.5
        ground_truth[:, 3] = scale * ground_truth[:, 3]

    return padded_image, ground_truth


def draw_bboxes(
    image: np.ndarray, pred_bboxes: np.ndarray, names: Dict[int, str]
):
    """
    @parma `image`:  Dim(height, width, channel)
    @parma `pred_bboxes`
        Dim(-1, (x, y, w, h, cls_id, prob))

    @return drawn_image

    Usage:
        image = yolo.draw_bboxes(image, bboxes)
    """
    image = np.copy(image)
    height, width, _ = image.shape

    bboxes = pred_bboxes * np.array([width, height, width, height, 1, 1])

    # Draw bboxes
    for bbox in bboxes:
        c_x = int(bbox[0])
        c_y = int(bbox[1])
        half_w = int(bbox[2] / 2)
        half_h = int(bbox[3] / 2)

        font_size = 0.4
        font_thickness = 1

        cls_id = int(bbox[4])
        prob = bbox[5]
        color = BBOX_COLORS[cls_id]

        if prob > 0.24:
            # Draw box
            top = c_x - half_w
            if top < 10:
                top = 10
            left = c_y - half_h
            if left < 0:
                left = 0
            bottom = c_x + half_w
            if bottom > height:
                bottom = height
            right = c_y + half_h
            if right > width:
                right = width

            top_left = (top, left)
            bottom_right = (c_x + half_w, c_y + half_h)
            cv2.rectangle(image, top_left, bottom_right, color, 2)

            # Draw text box
            bbox_text = "{}: {:.1%}".format(names[cls_id], prob)
            t_size = cv2.getTextSize(bbox_text, 0, font_size, font_thickness)[0]
            cv2.rectangle(
                image,
                top_left,
                (top + t_size[0], left - t_size[1] - 3),
                color,
                -1,
            )

            # Draw text
            cv2.putText(
                image,
                bbox_text,
                (top_left[0], top_left[1] - 2),
                cv2.FONT_HERSHEY_SIMPLEX,
                font_size,
                (
                    255 - color[0],
                    255 - color[1],
                    255 - color[2],
                ),
                font_thickness,
                lineType=cv2.LINE_AA,
            )

    return image



"""
MIT License

Copyright (c) 2021 Hyeonki Hong <hhk7734@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import time

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.callbacks import Callback

from ....common.config import YOLOConfig


class YOLOCallbackAtEachStep(Callback):
    """
    Ref
        - tf.keras.callbacks.LearningRateScheduler
    """

    def __init__(self, config: YOLOConfig, verbose: int):
        super().__init__()
        self._cfg_burn_in = config.net.burn_in
        self._cfg_learning_rate = config.net.learning_rate
        self._cfg_max_step = config.net.max_batches
        self._cfg_power = config.net.power
        self._cfg_scales = config.net.scales
        self._cfg_scale_steps = config.net.steps

        self._verbose = verbose

    def on_train_begin(self, logs=None):
        self.model._iou_loss = tf.Variable(0, dtype=tf.float32, trainable=False)
        self.model._obj_loss = tf.Variable(0, dtype=tf.float32, trainable=False)
        self.model._cls_loss = tf.Variable(0, dtype=tf.float32, trainable=False)
        self.model._total_loss = tf.Variable(
            0, dtype=tf.float32, trainable=False
        )

        self.model._total_truth = tf.Variable(
            0, dtype=tf.int64, trainable=False
        )
        self.model._ious = tf.Variable(0, dtype=tf.float32, trainable=False)
        self.model._recall50 = tf.Variable(0, dtype=tf.int32, trainable=False)
        self.model._recall75 = tf.Variable(0, dtype=tf.int32, trainable=False)

    def on_train_batch_begin(self, batch, logs=None):
        self.model._iou_loss.assign(0)
        self.model._obj_loss.assign(0)
        self.model._cls_loss.assign(0)
        self.model._total_loss.assign(0)

        self._prev_total_truth = self.model._total_truth.value()
        self.model._ious.assign(0)
        self.model._recall50.assign(0)
        self.model._recall75.assign(0)

        self.start_time = time.time()
        self.update_lr()

    def on_train_batch_end(self, batch, logs=None):
        # compile: steps_per_execution
        # next begin iteration number
        step = self.model._train_counter
        spe = self.model._steps_per_execution
        spe_f = tf.cast(spe, dtype=tf.float32)

        leraning_rate = K.get_value(self.model.optimizer.lr)

        iou_loss = self.model._iou_loss.value() / spe_f
        obj_loss = self.model._obj_loss.value() / spe_f
        cls_loss = self.model._cls_loss.value() / spe_f
        total_loss = self.model._total_loss.value() / spe_f

        total_truth = self.model._total_truth.value()
        truth = total_truth - self._prev_total_truth
        truth_f = tf.cast(truth, dtype=tf.float32)
        iou = self.model._ious.value() / truth_f
        recall50 = self.model._recall50.value()
        recall50_f = tf.cast(recall50, dtype=tf.float32) / truth_f
        recall75 = self.model._recall75.value()
        recall75_f = tf.cast(recall75, dtype=tf.float32) / truth_f

        tf.summary.scalar(name="loss/iou_loss", data=iou_loss, step=step)
        tf.summary.scalar(name="loss/obj_loss", data=obj_loss, step=step)
        tf.summary.scalar(name="loss/cls_loss", data=cls_loss, step=step)
        tf.summary.scalar(name="loss/total_loss", data=total_loss, step=step)

        tf.summary.scalar(
            name="metric/total_truth", data=total_truth, step=step
        )
        tf.summary.scalar(name="metric/truth", data=truth, step=step)
        tf.summary.scalar(name="metric/iou", data=iou, step=step)
        tf.summary.scalar(name="metric/recall50", data=recall50_f, step=step)
        tf.summary.scalar(name="metric/recall75", data=recall75_f, step=step)

        if self._verbose == 3:
            verbose = (
                f"step: {step.numpy()}({step.numpy()/self._cfg_max_step:.1%}), "
            )
            verbose += f"{time.time()-self.start_time:6.2f}s, "
            verbose += f"Truth: {truth.numpy():4}, "
            verbose += f"Avg IOU: {iou.numpy():5.3f}, "
            verbose += f".5R: {recall50_f.numpy():5.3f}, "
            verbose += f".75R: {recall75_f.numpy():5.3f}, "
            verbose += f"lr: {leraning_rate:7.5f}, "
            verbose += "Loss => "
            verbose += f"IOU: {iou_loss.numpy():6.3f}, "
            verbose += f"OBJ: {obj_loss.numpy():6.3f}, "
            verbose += f"CLS: {cls_loss.numpy():6.3f}, "
            verbose += f"Total: {total_loss.numpy():6.3f}"
            print(verbose)

        logs = logs or {}
        logs["lr"] = leraning_rate
        if step >= self._cfg_max_step:
            self.model.stop_training = True

    def update_lr(self):
        # on_train_batch_begin
        step = self.model._train_counter

        if not hasattr(self.model.optimizer, "lr"):
            raise ValueError('Optimizer must have a "lr" attribute.')

        lr = self._cfg_learning_rate

        # burn_in=1000
        # 0, 1, 2, ..., 999
        if step < self._cfg_burn_in:
            lr *= K.pow((step + 1) / self._cfg_burn_in, self._cfg_power)

        else:
            # scales=1600,1800
            # max_batches=2000
            for index, it in enumerate(
                [*self._cfg_scale_steps, self._cfg_max_step]
            ):
                if step < it:
                    for j in range(index):
                        lr *= self._cfg_scales[j]
                    break

        if lr != float(K.get_value(self.model.optimizer.lr)):
            tf.summary.scalar(name="learning_rate", data=lr, step=step)

        K.set_value(self.model.optimizer.lr, K.get_value(lr))
